--[[
  Copyright (c) 2021 by Plexim GmbH
  All rights reserved.

  A free license is granted to anyone to use this software for any legal
  non safety-critical purpose, including commercial applications, provided
  that:
  1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
  2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
--]] --
local Module = {}

local static = {numInstances = 0, instances = {}, finalized = nil}

-- private function
function DetermineBrp(baud, clkIn, maxClkDiv)
  local maxDiv = math.floor(clkIn / baud / 5) -- (how) should we round?
  local minDiv = math.floor(clkIn / baud / 13) -- (how) should we round?

  if minDiv > maxClkDiv then
    return nil
  end

  -- search for exact match of highest resolution
  local brp
  for div = minDiv, maxDiv do
    local fq = clkIn / baud / div
    if fq == math.floor(fq) then
      brp = div
      break
    end
  end
  return brp
end

-- private function
function DetermineTsegs(baud, clkin, brp, samplePoint)
  local seg = math.floor(clkin / brp / baud + 0.5)

  local tseg1 = math.floor(samplePoint * seg + 0.5) - 1
  if tseg1 < 1 then
    tseg1 = 1
  elseif tseg1 > 16 then
    tseg1 = 16
  end

  local tseg2 = seg - 1 - tseg1
  return tseg1, tseg2
end

-- private function
function CalculateCanTiming(settings)
  local timing = {}
  local cclk = settings['clk'] / settings['brp']
  timing['tq'] = (1 + settings['tseg1'] + settings['tseg2']) / cclk
  timing['baud'] = 1 / timing['tq']
  timing['sample'] = (1 + settings['tseg1']) /
                         (1 + settings['tseg1'] + settings['tseg2'])
  return timing
end

function Module.getBlock(globals)

  local Can = require('blocks.block').getBlock(globals)
  Can["instance"] = static.numInstances
  static.numInstances = static.numInstances + 1

  function Can:checkMaskParameters(env)
  end

  function Can:createImplicit(can)
    self.can = can
    static.instances[self.can] = self.bid

    self.can_letter = string.char(65 + self.can)
    self.mailboxes = {}
    self.num_mailboxes = 0
    self.is_configured = false

    self:logLine('CAN %s implicitly created.' % {self.can_letter})
  end

  function Can:configure(params, req)
    self.sample_point = params.sample_point
    self.baud = params.baud
    self.gpio = params.gpio
    self.auto_buson = params.auto_buson

    local canpins = '%s_GPIO%i_GPIO%i' %
                        {
          string.char(65 + self.can), Block.Mask.gpio[1], Block.Mask.gpio[2]
        }
    self.pinset =
        globals.target.getTargetParameters()['cans']['pin_sets'][canpins]

    if self.pinset == nil then
      return 'Pins %i/%i not supported for CAN %s.' %
                 {
            Block.Mask.gpio[1], Block.Mask.gpio[2], string.char(65 + self.can)
          }
    end

    self.clk, brpMax = globals.target.getCanClkAndMaxBrp()
    self.brp = DetermineBrp(self.baud, self.clk, brpMax)
    if self.brp == nil then
      return "CAN baud rate not achievable."
    end
    self.is_configured = true
  end

  function Can:getTxMailbox()
    if self.num_mailboxes == 32 then
      return 'Only 32 CAN mailboxes available.'
    end
    local mbox = self.num_mailboxes
    self.num_mailboxes = self.num_mailboxes + 1
    return mbox
  end

  function Can:setupTxMailbox(mbox, params)
    self.mailboxes[mbox] = {
      is_tx = true,
      can_id = params.can_id,
      ext_id = params.ext_id,
      width = params.width
    }
  end

  function Can:getRxMailbox()
    if self.num_mailboxes == 16 then
      return 'Out of available CAN mailboxes.'
    end
    local mbox = self.num_mailboxes
    self.num_mailboxes = self.num_mailboxes + 1
    return mbox
  end

  function Can:setupRxMailbox(mbox, params)
    self.mailboxes[mbox] = {
      is_tx = false,
      can_id = params.can_id,
      ext_id = params.ext_id,
      width = params.width
    }
  end

  function Can:getDirectFeedthroughCode()
    return "Explicit use of CAN via target block not supported."
  end

  function Can:getNonDirectFeedthroughCode()
    return "Explicit use of CAN via target block not supported."
  end

  function Can:finalizeThis(c)
    self.tseg1, self.tseg2 = DetermineTsegs(self.baud, self.clk, self.brp,
                                            self.sample_point)

    -- make synchronization jump width as large as possible
    self.sjw = 4
    if self.sjw > self.tseg2 then
      self.sjw = self.tseg2
    end

    self.timing = CalculateCanTiming({
      brp = self.brp,
      clk = self.clk,
      tseg1 = self.tseg1,
      tseg2 = self.tseg2
    })

    self:logLine('CAN settings: %s' % {dump(self.timing)})

    c.PreInitCode:append(
        "// Configure CAN %s at %.3f Bit/s, with sampling at %.1f%%" %
            {
              string.char(65 + self.can), self.timing['baud'],
              self.timing['sample'] * 100
            })
    c.PreInitCode:append("{")
    c.PreInitCode:append("  PLX_CANBUS_Params_t params;")
    c.PreInitCode:append("  params.tseg1 = %i;" % {self.tseg1})
    c.PreInitCode:append("  params.tseg2 = %i;" % {self.tseg2})
    c.PreInitCode:append("  params.sjw = %i;" % {self.sjw})
    c.PreInitCode:append("  params.sam = %i;" % {0}) -- 1x ((3x sampling not supported on newer MCUs)
    c.PreInitCode:append("  params.brp = %i;" % {self.brp})
    c.PreInitCode:append("  params.autoBusOn = %i;" % {self.auto_buson})
    c.PreInitCode:append(
        "  PLX_CANBUS_configure(CanHandles[%i], PLX_CANBUS_CAN_%s, %i, &params);" %
            {self.instance, string.char(65 + self.can), self.pinset})
    for mbox, params in pairs(self.mailboxes) do
      c.PreInitCode:append(
          "  (void)PLX_CANBUS_setupMailbox(CanHandles[%i], %i, %i, %i, %i, %i);" %
              {
                self.instance, mbox, params.is_tx, params.can_id, params.ext_id,
                params.width
              })
    end
    c.PreInitCode:append("}")
    return c
  end

  function Can:finalize(c)
    if static.finalized ~= nil then
      return {}
    end

    c.Include:append('plx_canbus.h')
    c.Declarations:append('PLX_CANBUS_Handle_t CanHandles[%i];' %
                              {static.numInstances})
    c.Declarations:append('PLX_CANBUS_Obj_t CanObj[%i];' % {static.numInstances})

    c.Declarations:append(
        'bool PLXHAL_CAN_getMessage(uint16_t aChannel, uint16_t aMailBox, unsigned char data[], unsigned char lenMax){')
    c.Declarations:append(
        '  return PLX_CANBUS_getMessage(CanHandles[aChannel], aMailBox, data, lenMax);')
    c.Declarations:append('}')

    c.Declarations:append(
        'void PLXHAL_CAN_putMessage(uint16_t aChannel, uint16_t aMailBox, unsigned char data[], unsigned char len){')
    c.Declarations:append(
        '  (void)PLX_CANBUS_putMessage(CanHandles[aChannel], aMailBox, data, len);')
    c.Declarations:append('}')

    c.Declarations:append(
        'void PLXHAL_CAN_setBusOn(uint16_t aChannel, bool aBusOn){')
    c.Declarations:append('  PLX_CANBUS_setBusOn(CanHandles[aChannel], aBusOn);')
    c.Declarations:append('}')

    c.Declarations:append('bool PLXHAL_CAN_getIsBusOn(uint16_t aChannel){')
    c.Declarations:append('  return PLX_CANBUS_isBusOn(CanHandles[aChannel]);')
    c.Declarations:append('}')

    c.Declarations:append('bool PLXHAL_CAN_getIsErrorActive(uint16_t aChannel){')
    c.Declarations:append(
        '  return PLX_CANBUS_isErrorActive(CanHandles[aChannel]);')
    c.Declarations:append('}')

    local code = [[
    {
      PLX_CANBUS_sinit();
      int i;
      for(i=0; i<%d; i++)
      {
        CanHandles[i] = PLX_CANBUS_init(&CanObj[i], sizeof(CanObj[i]));
      }
    }]]
    c.PreInitCode:append(code % {static.numInstances})

    for _, bid in pairs(static.instances) do
      local can = globals.instances[bid]
      local c = can:finalizeThis(c)
      if type(c) == 'string' then
        return c
      end
    end

    static.finalized = true
    return c
  end

  return Can
end

return Module
